/*
  Top10, a racing simulator
  Copyright (C) 2000-2005  Johann Deneux
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  Authors can be contacted at following electronic addresses:
  Johann Deneux: johann.deneux@it.uu.se
*/

#include "CheckpointsDialog.hh"
#include "DialogManager.hh"
#include <cassert>
#include <sstream>

/*
 * Callbacks
 */
/* This is a template callback. The template parameters are the affected point  (1 or 2),
   and the component (0:X, 1:Y, Z:2)
*/
template<int pointN, int coordN>
static void inputCB(puObject* ob)
{
  assert(pointN == 1 || pointN == 2);
  assert(coordN >= 0 && coordN < 3);

  CheckpointsDialog* dialog = (CheckpointsDialog*)ob->getUserData();

  int idx = dialog->getManager()->cps_ed->getCurrentIndex();
  if (idx == -1) return;

  // Get the point
  top10::math::Vector pt;
  if (pointN == 2) pt = dialog->getManager()->cps_ed->getCurrent().getPoint2();
  else if (pointN == 1) pt = dialog->getManager()->cps_ed->getCurrent().getPoint1();
  else assert(false);

  // Retrieve the input's value and modify the component
  float tmp;
  ob->getValue(&tmp);
  pt[coordN] = tmp;

  // Set the point
  if (pointN == 2) dialog->getManager()->cps_ed->checkpoints.at(idx).setPoint2(pt);
  else if (pointN == 1) dialog->getManager()->cps_ed->checkpoints.at(idx).setPoint1(pt);  
}

template<int pointN>
static void pickCB(puObject* ob)
{
  assert(pointN == 1 || pointN == 2);
  CheckpointsDialog* dialog = (CheckpointsDialog*)ob->getUserData();

  if (dialog->getManager()->cps_ed->getCurrentIndex() == -1) return;

  top10::math::Vector pt;
  if (pointN == 1)
    dialog->getManager()->cps_ed->pickLeft();
  else if (pointN == 2)
    dialog->getManager()->cps_ed->pickRight();
  else assert(false);

  dialog->remove();
  dialog->install();
  dialog->show();
}

static void nextCB(puObject* ob)
{
  CheckpointsDialog* dialog = (CheckpointsDialog*)ob->getUserData();
  try {
    dialog->getManager()->cps_ed->next();
  }
  catch(std::string e) {
    std::cout<<e<<std::endl;
  }

  dialog->remove();
  dialog->install();
  dialog->show();
}

static void prevCB(puObject* ob)
{
  CheckpointsDialog* dialog = (CheckpointsDialog*)ob->getUserData();
  try {
    dialog->getManager()->cps_ed->previous();
  }
  catch(std::string e) {
    std::cout<<e<<std::endl;
  }

  dialog->remove();
  dialog->install();
  dialog->show();
}

static void addCB(puObject* ob)
{
  CheckpointsDialog* dialog = (CheckpointsDialog*)ob->getUserData();
  dialog->getManager()->cps_ed->addNew();

  dialog->remove();
  dialog->install();
  dialog->show();
}

static void delCB(puObject* ob)
{
  CheckpointsDialog* dialog = (CheckpointsDialog*)ob->getUserData();
  dialog->getManager()->cps_ed->remove();

  dialog->remove();
  dialog->install();
  dialog->show();
}

/*
 * Class: CheckpointsDialog
 */
CheckpointsDialog::CheckpointsDialog(DialogManager* manager):
  Dialog(manager, "Checkpoints")
{
}

void CheckpointsDialog::populate()
{
  top10::math::Vector pt = getManager()->cps_ed->getCurrent().getPoint2();

  puInput* pt2_x = new puInput(0, H, M_W, 2*H);
  pt2_x->setUserData(this);
  pt2_x->setCallback(inputCB<2, 0>);
  pt2_x->setValue((float)pt.x);

  puInput* pt2_y = new puInput(0, 2*H, M_W, 3*H);
  pt2_y->setUserData(this);
  pt2_y->setCallback(inputCB<2, 1>);
  pt2_y->setValue((float)pt.y);

  puInput* pt2_z = new puInput(0, 3*H, M_W, 4*H);
  pt2_z->setUserData(this);
  pt2_z->setCallback(inputCB<2, 2>);
  pt2_z->setValue((float)pt.z);

  pt = getManager()->cps_ed->getCurrent().getPoint1();
  puInput* pt1_x = new puInput(0, 4.5*H, M_W, 5.5*H);
  pt1_x->setUserData(this);
  pt1_x->setCallback(inputCB<1, 0>);
  pt1_x->setValue((float)pt.x);

  puInput* pt1_y = new puInput(0, 5.5*H, M_W, 6.5*H);
  pt1_y->setUserData(this);
  pt1_y->setCallback(inputCB<1, 1>);
  pt1_y->setValue((float)pt.y);

  puInput* pt1_z = new puInput(0, 6.5*H, M_W, 7.5*H);
  pt1_z->setUserData(this);
  pt1_z->setCallback(inputCB<1, 2>);
  pt1_z->setValue((float)pt.z);

  puOneShot* pick2 = new puOneShot(M_W +5, H, M_W+5 +M_W, 4*H);
  pick2->setLegend("Pick");
  pick2->setUserData(this);
  pick2->setCallback(pickCB<2>);

  puOneShot* pick1 = new puOneShot(M_W +5, 4.5*H, M_W+5 +M_W, 7.5*H);
  pick1->setLegend("Pick");
  pick1->setUserData(this);
  pick1->setCallback(pickCB<1>);

  puOneShot* prev_shot = new puOneShot(0, 8*H, S_W, 9*H);
  prev_shot->setUserData(this);
  prev_shot->setCallback(prevCB);
  prev_shot->setLegend("<");

  puOneShot* next_shot = new puOneShot(2*S_W, 8*H, 3*S_W, 9*H);
  next_shot->setUserData(this);
  next_shot->setCallback(nextCB);
  next_shot->setLegend(">");

  puOneShot* add_shot = new puOneShot(4*S_W, 8*H, 5*S_W, 9*H);
  add_shot->setUserData(this);
  add_shot->setCallback(addCB);
  add_shot->setLegend("+");

  puOneShot* del_shot = new puOneShot(6*S_W, 8*H, 7*S_W, 9*H);
  del_shot->setUserData(this);
  del_shot->setCallback(delCB);
  del_shot->setLegend("-");

  std::ostringstream tmp;
  tmp<<getManager()->cps_ed->getCurrentIndex();
  puText* idx_text = new puText(S_W, 8*H);
  idx_str = tmp.str();
  idx_text->setLabel(idx_str.c_str());
}
